;;; develock.el --- additional font-lock keywords for the developers

;; Copyright (C) 2001, 2002 Katsumi Yamaoka
;; Copyright (C) 2001, 2002 Jun'ichi Shiono

;; Author: Katsumi Yamaoka <yamaoka@jpl.org>
;;         Jun'ichi Shiono <jun@fsas.fujitsu.com>
;; Created: 2001/06/28
;; Revised: 2002/02/04
;; Keywords: font-lock emacs-lisp change-log texinfo c java perl mail news

;; Develock is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; Develock is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with Develock; see the file COPYING.  If not, write to the
;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Commentary:

;; Develock is a minor mode which provides the ability to make font-
;; lock highlight leading and trailing whitespaces, long lines and the
;; other garbages in the file buffer for the Lisp modes, the Change
;; Log mode, the Texinfo mode, the C modes, the Java mode, the CPerl
;; mode, the Perl mode and some Mail modes.  Here is an example of how
;; to set up your startup file (possibly .emacs) to use Develock:
;;
;;(cond ((featurep 'xemacs)
;;       (require 'develock)
;;       ;; `turn-on-develock' is equivalent to `turn-on-font-lock',
;;       ;;  except that it does not highlight the startup screen.
;;       (add-hook 'lisp-interaction-mode-hook 'turn-on-develock)
;;       (add-hook 'mail-setup-hook 'turn-on-font-lock))
;;      ((>= emacs-major-version 20)
;;       (require 'develock)
;;       (add-hook 'lisp-interaction-mode-hook 'turn-on-font-lock)
;;       (add-hook 'emacs-lisp-mode-hook 'turn-on-font-lock)
;;       (add-hook 'change-log-mode-hook 'turn-on-font-lock)
;;       (add-hook 'texinfo-mode-hook 'turn-on-font-lock)
;;       (add-hook 'c-mode-common-hook 'turn-on-font-lock)
;;       (add-hook 'cperl-mode-hook 'turn-on-font-lock)
;;       (add-hook 'perl-mode-hook 'turn-on-font-lock)
;;       (add-hook 'mail-setup-hook 'turn-on-font-lock)
;;       (add-hook 'message-mode-hook 'turn-on-font-lock)))
;;
;; Note that Develock might be useless if you set the value of the
;; variable `indent-tabs-mode' to nil.
;;
;; Develock will work with Emacs 20.4 and later and XEmacs 21.1 and
;; later.  Needless to say, you have to install this file as follows:
;;
;; # cp -p develock.el /usr/local/share/emacs/site-lisp
;; # cd /usr/local/share/emacs/site-lisp
;; # emacs -q -no-site-file -batch -f batch-byte-compile develock.el
;;
;; It is recommended that you use the demand-driven fontification tool
;; (e.g. jit-lock, lazy-shot) together with font-lock.
;;
;; You may find latest version of Develock in ftp.jpl.org:/pub/elisp/.

;; Texinfo
;; =======
;; The value of `fill-column' for `texinfo-mode' has changed to 70 in
;; Emacs 21.  However, many texinfo files are written on the condition
;; that the value is 72.  If too much highlighting offends your eyes,
;; you can change the behavior of highlighting as shown below:
;;
;;(if (and (not (featurep 'xemacs))
;;	 (>= emacs-major-version 21))
;;    (plist-put develock-max-column-plist 'texinfo-mode 72))

;; Ediff
;; =====
;; When you run Ediff on the Develock'ed buffers, you may feel
;; everything is in confusion.  For such a case, the following hooks
;; may help you see diffs clearly.
;;
;;(add-hook
;; 'ediff-prepare-buffer-hook
;; (lambda nil
;;   (if (and (boundp 'font-lock-mode) font-lock-mode
;;	    (boundp 'develock-mode) develock-mode)
;;       (progn
;;	 (develock-mode 0)
;;	 (set (make-local-variable 'develock-mode-suspended) t)))))
;;
;;(add-hook
;; 'ediff-cleanup-hook
;; (lambda nil
;;   (let ((buffers (list ediff-buffer-A ediff-buffer-B ediff-buffer-C)))
;;     (save-excursion
;;       (while buffers
;;	 (if (buffer-live-p (car buffers))
;;	     (progn
;;	       (set-buffer (car buffers))
;;	       (if (and (boundp 'develock-mode-suspended)
;;			develock-mode-suspended)
;;		   (progn
;;		     (develock-mode 1)
;;		     (makunbound 'develock-mode-suspended)))))
;;	 (setq buffers (cdr buffers)))))))

;; Wanderlust
;; ==========
;; Though Wanderlust does not support font-lock, you may use Develock
;; for `wl-draft-mode'.  Try to use the following codes in .wl file.
;;
;;(require 'develock)
;;(require 'message)
;;
;;(setq wl-highlight-body-too nil)
;;
;;(defadvice wl-highlight-headers (around dont-highlight-draft
;;					(&optional for-draft) activate)
;;  "Don't highlight draft buffers."
;;  (or for-draft ad-do-it))
;;
;;(defun wl-draft-setup-develock ()
;;  "Function used to setup to use Develock in `wl-draft-mode'."
;;  (if (featurep 'xemacs)
;;      (put 'wl-draft-mode 'font-lock-defaults
;;	   '(message-font-lock-keywords t))
;;    (make-local-variable 'font-lock-defaults)
;;    (setq font-lock-defaults '(message-font-lock-keywords t t)))
;;  (turn-on-font-lock))
;;
;;(add-hook 'wl-mail-setup-hook 'wl-draft-setup-develock)
;;
;;(setq develock-max-column-plist
;;      (plist-put develock-max-column-plist 'wl-draft-mode t))
;;
;;(let ((elem (copy-sequence (assq 'message-mode develock-keywords-alist))))
;;  (setcar elem 'wl-draft-mode)
;;  (setq develock-keywords-alist
;;	(cons elem (delq (assq 'wl-draft-mode develock-keywords-alist)
;;			 develock-keywords-alist))))
;;
;; If you feel that highlighting is insufficient in gaudiness,
;; customize the value of the variable `message-font-lock-keywords' to
;; make font-lock do what you want it to.

;; Mew
;; ===
;; Similarly, you can use Develock for `mew-draft-mode'.  The following
;; codes are for .mew file.  You will never know unless you try.
;;
;;(require 'develock)
;;(require 'message)
;;
;;(defadvice mew-highlight-header-region (around dont-highlight-draft activate)
;;  "Don't highlight draft buffers."
;;  (or (eq 'mew-draft-mode major-mode) ad-do-it))
;;
;;(defadvice mew-highlight-body-region (around dont-highlight-draft activate)
;;  "Don't highlight draft buffers."
;;  (or (eq 'mew-draft-mode major-mode) ad-do-it))
;;
;;(defun mew-draft-setup-develock ()
;;  "Function used to setup to use Develock in `mew-draft-mode'."
;;  (if (featurep 'xemacs)
;;      (put 'mew-draft-mode 'font-lock-defaults
;;	   '(message-font-lock-keywords t))
;;    (make-local-variable 'font-lock-defaults)
;;    (setq font-lock-defaults '(message-font-lock-keywords t t)))
;;  (turn-on-font-lock))
;;
;;(add-hook 'mew-draft-mode-hook 'mew-draft-setup-develock)
;;
;;(setq develock-max-column-plist
;;      (plist-put develock-max-column-plist 'mew-draft-mode t))
;;
;;(let ((elem (copy-sequence (assq 'message-mode develock-keywords-alist))))
;;  (setcar elem 'mew-draft-mode)
;;  (setq develock-keywords-alist
;;	(cons elem (delq (assq 'mew-draft-mode develock-keywords-alist)
;;			 develock-keywords-alist))))

;;; Code:

(defconst develock-version "0.9"
  "Version number for this version of Develock.")

(require 'font-lock)

(defgroup develock nil
  "A tool for the developers."
  :group 'development)

(make-variable-buffer-local
 (defvar develock-mode nil
   "Internal buffer-local variable used to indicate whether Develock is on."))

(defcustom develock-mode-strings '(" Develock" " Font")
  "List of identifications to put in the modeline whether Develock mode
is on or off.  Each element should be a string beginning with a space,
a null string or a symbol with a similar string as its value.  You can
use faces for each element if you are using Emacs 21 as follows:

\(let* ((on (copy-face 'mode-line 'develock-mode-on-face))
       (off (copy-face on 'develock-mode-off-face)))
  (set-face-foreground on \"PaleTurquoise\")
  (set-face-foreground off \"MistyRose\")
  (setq develock-mode-strings (list (propertize \" Develock\" 'face on)
				    (propertize \" Font\" 'face off)))
  (add-minor-mode 'font-lock-mode
		  (cons 'develock-mode develock-mode-strings)))

Otherwise if you are using XEmacs, you can use a face for all the
minor mode strings as shown below:

\(copy-face 'modeline 'modeline-mousable-minor-mode)
\(set-face-foreground 'modeline-mousable-minor-mode \"PaleTurquoise\")
"
  :type '(list (sexp :tag "On") (sexp :tag "Off"))
  :set (lambda (symbol value)
	 (prog1
	     (set-default symbol value)
	   (let ((id (cons 'develock-mode value))
		 mmode)
	     (if (fboundp 'add-minor-mode)
		 (add-minor-mode 'font-lock-mode id)
	       (if (setq mmode (assq 'font-lock-mode minor-mode-alist))
		   (setcdr mmode (list id))
		 (setq minor-mode-alist (cons (list 'font-lock-mode id)
					      minor-mode-alist)))))))
  :group 'develock
  :group 'font-lock)

(if (fboundp 'add-minor-mode)
    (eval-after-load "font-lock"
      '(add-minor-mode 'font-lock-mode
		       (cons 'develock-mode develock-mode-strings)))
  (eval-after-load "font-lock"
    '(setcdr (assq 'font-lock-mode minor-mode-alist)
	     (list (cons 'develock-mode develock-mode-strings)))))

(defface develock-whitespaces-face-1
  '((t (:background "Red")))
  "Develock face used to highlight whitespaces."
  :group 'develock
  :group 'font-lock-highlighting-faces
  :group 'font-lock-faces)

(defface develock-whitespaces-face-2
  '((t (:background "Orange")))
  "Develock face used to highlight whitespaces."
  :group 'develock
  :group 'font-lock-highlighting-faces
  :group 'font-lock-faces)

(defface develock-whitespaces-face-3
  '((t (:background "Yellow")))
  "Develock face used to highlight whitespaces."
  :group 'develock
  :group 'font-lock-highlighting-faces
  :group 'font-lock-faces)

(defface develock-long-line-face-1
  '((((background dark))
     (:foreground "Orange"))
    (t
     (:foreground "Red")))
  "Develock face used to highlight long lines."
  :group 'develock
  :group 'font-lock-highlighting-faces
  :group 'font-lock-faces)

(defface develock-long-line-face-2
  '((t (:foreground "Red" :background "Yellow")))
  "Develock face used to highlight long lines."
  :group 'develock
  :group 'font-lock-highlighting-faces
  :group 'font-lock-faces)

(defface develock-lonely-parentheses-face
  '((t (:foreground "Blue" :background "PaleTurquoise")))
  "Develock face used to highlight lonely parentheses."
  :group 'develock
  :group 'font-lock-highlighting-faces
  :group 'font-lock-faces)

(defface develock-bad-manner-face
  '((((background dark))
     (:foreground "Black" :background "Yellow"))
    (t
     (:foreground "Yellow" :background "Black")))
  "Develock face used to highlight characters (or words) which violate
the manners of mail messages or news articles."
  :group 'develock
  :group 'font-lock-highlighting-faces
  :group 'font-lock-faces)

(defface develock-reachable-mail-address-face
  '((t (:foreground "DarkGreen" :background "LemonChiffon")))
  "Develock face used to highlight reachable E-mail addresses.  That
would be defenseless to SPAMMERS.")

(let ((faces '(develock-whitespaces-face-1
	       develock-whitespaces-face-2
	       develock-whitespaces-face-3
	       develock-long-line-face-1
	       develock-long-line-face-2
	       develock-lonely-parentheses-face
	       develock-bad-manner-face
	       develock-reachable-mail-address-face)))
  (while faces
    (set (car faces) (car faces))
    (setq faces (cdr faces))))

(defcustom develock-max-column-plist (list 'emacs-lisp-mode 79
					   'lisp-interaction-mode 'w
					   'change-log-mode t
					   'texinfo-mode t
					   'c-mode 79
					   'c++-mode 79
					   'java-mode 79
					   'cperl-mode 79
					   'perl-mode 79
					   'mail-mode t
					   'message-mode t
					   'cmail-mail-mode t)
  "Plist of `major-mode' and a limitation that lines should not be longer
than the value.  Value `w' means one subtracted from the window width.
Value t means the return value of `current-fill-column'.  You can
inhibit to highlight long lines using the value nil."
  :type '(repeat (list :inline t :format "%v"
		       (symbol :tag "Major-mode")
		       (choice :tag "Value"
			       (integer :tag "Integer")
			       (const :tag "Window_width -1" w)
			       (const :tag "Fill_column" t)
			       (const :tag "Off" nil))))
  :group 'develock
  :group 'font-lock)

(defcustom develock-auto-enable t
  "If non-nil, turn Develock mode on when font-lock is turned on."
  :type 'boolean
  :group 'develock
  :group 'font-lock)

(defcustom develock-mode-ignore-kinsoku-list
  '(emacs-lisp-mode lisp-interaction-mode c-mode c++-mode java-mode
		    cperl-mode perl-mode)
  "List of `major-mode' to ignore kinsoku in the end of each line."
  :type '(repeat (symbol :tag "Major-mode"))
  :group 'develock
  :group 'font-lock)

(eval-and-compile
  (if (featurep 'xemacs)
      (progn
	(defalias 'develock-point-at-bol 'point-at-bol)
	(defalias 'develock-point-at-eol 'point-at-eol))
    (defalias 'develock-point-at-bol 'line-beginning-position)
    (defalias 'develock-point-at-eol 'line-end-position)))

(defvar develock-lisp-font-lock-keywords
  '(;; a long line
    (develock-find-long-lines
     (1 develock-long-line-face-1 t)
     (2 develock-long-line-face-2 t))
    ;; long spaces
    ("\\(\t*\\)\\(        +\\)"
     (1 develock-whitespaces-face-2)
     (2 develock-whitespaces-face-3))
    ;; trailing whitespaces
    ("[^\t\n ]\\([\t ]+\\)$"
     (1 develock-whitespaces-face-1 t))
    ;; spaces before tabs
    ("\\( +\\)\\(\t+\\)"
     (1 develock-whitespaces-face-1 t)
     (2 develock-whitespaces-face-2 t))
    ;; tab space tab
    ("\\(\t\\) \t"
     (1 develock-whitespaces-face-2 append))
    ;; only tabs or spaces in the line
    ("^[\t ]+$"
     (0 develock-whitespaces-face-2 append))
    ;; lonely left parentheses or brackets without a comment
    ("^[\t ]*\\(['`]?[\t ([]*[([]\\)[\t ]*$"
     1 develock-lonely-parentheses-face)
    ;; lonely right parentheses or brackets
    ("^[\t ]*\\([]\t )]*[])]\\)\\([\t ;]+\\|[\t ]*$\\)"
     1 develock-lonely-parentheses-face)
    ;; whitespaces after a left parenthesis
    ("(\\([\t ]+\\)[^\n;]"
     1 develock-whitespaces-face-2)
    ;; whitespaces before a right parenthesis
    ("([^\n]\\([\t ]+\\))"
     1 develock-whitespaces-face-2)
    ;; reachable E-mail addresses
    ("<?[-.0-9A-Z_a-z]+@[-0-9A-Z_a-z]+\\(\\.[-0-9A-Z_a-z]+\\)+>?"
     (0 develock-reachable-mail-address-face t)))
  "*Extraordinary level highlighting for the Lisp modes.")

(defvar develock-change-log-font-lock-keywords
  '(;; a long line
    (develock-find-long-lines
     (1 develock-long-line-face-1 append)
     (2 develock-long-line-face-2 append))
    ;; tabs or 3 or more spaces in the entry line
    ("\t[\t ]*\\|   +"
     (0 (if (memq (char-after (develock-point-at-bol)) '(?1 ?2))
	    develock-whitespaces-face-2)
	append))
    ;; a leading tab in the entry line
    ("^\\(\t\\)[12][0-9][0-9][0-9]-[01][0-9]-[0-3][0-9]  "
     (1 develock-whitespaces-face-2))
    ;; leading spaces
    ("^\\( +\\)[^\t\n]"
     (1 develock-whitespaces-face-3))
    ;; leading 2 or more tabs
    ("^\\([\t ][\t ]+\\)[^\t\n]"
     (1 develock-whitespaces-face-2))
    ;; trailing whitespaces
    ("[^\t\n ]\\([\t ]+\\)$"
     (1 develock-whitespaces-face-1 t))
    ;; tabs or 2 or more spaces in the log line
    ("[^\t\n ]\\(\t[\t ]*\\|  +\\)[^\t ]"
     (1 (if (and (memq (char-after (develock-point-at-bol)) '(?\t ?\ ))
		 (not (string-equal ".  "
				    (buffer-substring (1- (match-beginning 1))
						      (match-end 1)))))
	    develock-whitespaces-face-2)
	prepend))
    ;; spaces before tabs
    ("\\( +\\)\\(\t+\\)"
     (1 develock-whitespaces-face-1 t)
     (2 develock-whitespaces-face-2 append))
    ;; only tabs or spaces in the line
    ("^[\t ]+$"
     (0 develock-whitespaces-face-2 append))
    ;; whitespaces between a file name and a colon
    ("^\t\\* [^\t\n ]+\\( (.+)\\)?\\([\t ]+\\):"
     (2 develock-whitespaces-face-1 t)))
  "*Extraordinary level highlighting for the Change Log mode.")

(defvar develock-texinfo-font-lock-keywords
  '(;; a long line
    (develock-find-long-lines
     (1 develock-long-line-face-1 t)
     (2 develock-long-line-face-2 t))
    ;; trailing whitespaces
    ("[^\t\n ]\\([\t ]+\\)$"
     (1 develock-whitespaces-face-1 t))
    ;; spaces before tabs
    ("\\( +\\)\\(\t+\\)"
     (1 develock-whitespaces-face-1 t)
     (2 develock-whitespaces-face-2 t))
    ;; tab space tab
    ("\\(\t\\) \t"
     (1 develock-whitespaces-face-2 append))
    ;; only tabs or spaces in the line
    ("^[\t ]+$"
     (0 develock-whitespaces-face-2 append)))
  "*Extraordinary level highlighting for the TeXinfo mode.")

(defvar develock-c-font-lock-keywords
  '(;; a long line
    (develock-find-long-lines
     (1 develock-long-line-face-1 t)
     (2 develock-long-line-face-2 t))
    ;; long spaces
    ("\\(\t*\\)\\(        +\\)"
     (1 develock-whitespaces-face-2)
     (2 develock-whitespaces-face-3))
    ;; trailing whitespaces
    ("[^\t\n ]\\([\t ]+\\)$"
     (1 develock-whitespaces-face-1 t))
    ;; spaces before tabs
    ("\\( +\\)\\(\t+\\)"
     (1 develock-whitespaces-face-1 t)
     (2 develock-whitespaces-face-2 t))
    ;; tab space tab
    ("\\(\t\\) \t"
     (1 develock-whitespaces-face-2 append))
    ;; only tabs or spaces in the line
    ("^[\t ]+$"
     (0 develock-whitespaces-face-2 append))
    ;; reachable E-mail addresses
    ("<?[-.0-9A-Z_a-z]+@[-0-9A-Z_a-z]+\\(\\.[-0-9A-Z_a-z]+\\)+>?"
     (0 develock-reachable-mail-address-face t)))
  "*Extraordinary level highlighting for the C modes.")

(defvar develock-java-font-lock-keywords
  '(;; a long line
    (develock-find-long-lines
     (1 develock-long-line-face-1 t)
     (2 develock-long-line-face-2 t))
    ;; long spaces
    ("\\(\t*\\)\\(        +\\)"
     (1 develock-whitespaces-face-2)
     (2 develock-whitespaces-face-3))
    ;; trailing whitespaces
    ("[^\t\n ]\\([\t ]+\\)$"
     (1 develock-whitespaces-face-1 t))
    ;; spaces before tabs
    ("\\( +\\)\\(\t+\\)"
     (1 develock-whitespaces-face-1 t)
     (2 develock-whitespaces-face-2 t))
    ;; tab space tab
    ("\\(\t\\) \t"
     (1 develock-whitespaces-face-2 append))
    ;; only tabs or spaces in the line
    ("^[\t ]+$"
     (0 develock-whitespaces-face-2 append))
    ;; reachable E-mail addresses
    ("<?[-.0-9A-Z_a-z]+@[-0-9A-Z_a-z]+\\(\\.[-0-9A-Z_a-z]+\\)+>?"
     (0 develock-reachable-mail-address-face t)))
  "*Extraordinary level highlighting for the Java mode.")

(defvar develock-cperl-font-lock-keywords
  '(;; a long line
    (develock-find-long-lines
     (1 develock-long-line-face-1 t)
     (2 develock-long-line-face-2 t))
    ;; long spaces
    ("\\(\t*\\)\\(        +\\)"
     (1 develock-whitespaces-face-2)
     (2 develock-whitespaces-face-3))
    ;; trailing whitespaces
    ("[^\t\n ]\\([\t ]+\\)$"
     (1 develock-whitespaces-face-1 t))
    ;; spaces before tabs
    ("\\( +\\)\\(\t+\\)"
     (1 develock-whitespaces-face-1 t)
     (2 develock-whitespaces-face-2 t))
    ;; tab space tab
    ("\\(\t\\) \t"
     (1 develock-whitespaces-face-2 append))
    ;; only tabs or spaces in the line
    ("^[\t ]+$"
     (0 develock-whitespaces-face-2 append))
    ;; reachable E-mail addresses
    ("<?[-.0-9A-Z_a-z]+@[-0-9A-Z_a-z]+\\(\\.[-0-9A-Z_a-z]+\\)+>?"
     (0 develock-reachable-mail-address-face t)))
  "*Extraordinary level highlighting for the CPerl mode.")

(defvar develock-perl-font-lock-keywords
  '(;; a long line
    (develock-find-long-lines
     (1 develock-long-line-face-1 t)
     (2 develock-long-line-face-2 t))
    ;; long spaces
    ("\\(\t*\\)\\(        +\\)"
     (1 develock-whitespaces-face-2)
     (2 develock-whitespaces-face-3))
    ;; trailing whitespaces
    ("[^\t\n ]\\([\t ]+\\)$"
     (1 develock-whitespaces-face-1 t))
    ;; spaces before tabs
    ("\\( +\\)\\(\t+\\)"
     (1 develock-whitespaces-face-1 t)
     (2 develock-whitespaces-face-2 t))
    ;; tab space tab
    ("\\(\t\\) \t"
     (1 develock-whitespaces-face-2 append))
    ;; only tabs or spaces in the line
    ("^[\t ]+$"
     (0 develock-whitespaces-face-2 append))
    ;; reachable E-mail addresses
    ("<?[-.0-9A-Z_a-z]+@[-0-9A-Z_a-z]+\\(\\.[-0-9A-Z_a-z]+\\)+>?"
     (0 develock-reachable-mail-address-face t)))
  "*Extraordinary level highlighting for the Perl mode.")

(defvar develock-mail-font-lock-keywords
  `(;; a long line
    (develock-find-long-lines
     (1 develock-long-line-face-1 t)
     (2 develock-long-line-face-2 t))
    ;; long spaces
    ("\\(\t*\\)\\(        +\\)"
     (1 develock-whitespaces-face-2)
     (2 develock-whitespaces-face-3))
    ;; trailing whitespaces
    ("[^\t\n ]\\([\t ]+\\)$"
     (1 develock-whitespaces-face-1 t))
    ;; spaces before tabs
    ("\\( +\\)\\(\t+\\)"
     (1 develock-whitespaces-face-1 t)
     (2 develock-whitespaces-face-2 t))
    ;; tab space tab
    ("\\(\t\\) \t"
     (1 develock-whitespaces-face-2 append))
    ;; only tabs or spaces in the line
    ("^[\t ]+$"
     (0 develock-whitespaces-face-2 append))
    ,(if (featurep 'mule)
	 ;; Japanese hankaku katakana
	 '("\\ck+" (0 develock-bad-manner-face t))))
  "*Extraordinary level highlighting for the Mail modes.")

(defvar lisp-font-lock-keywords-x nil
  "Internal variable used to have extraordinary level font-lock keywords
for the Lisp modes.")

(defvar change-log-font-lock-keywords-x nil
  "Internal variable used to have extraordinary level font-lock keywords
for the Change Log mode.")

(defvar texinfo-font-lock-keywords-x nil
  "Internal variable used to have extraordinary level font-lock keywords
for the TeXinfo mode.")

(defvar c-font-lock-keywords-x nil
  "Internal variable used to have extraordinary level font-lock keywords
for the C modes.")

(defvar java-font-lock-keywords-x nil
  "Internal variable used to have extraordinary level font-lock keywords
for the Java mode.")

(defvar cperl-font-lock-keywords-x nil
  "Internal variable used to have extraordinary level font-lock keywords
for the CPerl mode.")

(defvar perl-font-lock-keywords-x nil
  "Internal variable used to have extraordinary level font-lock keywords
for the Perl mode.")

(defvar mail-font-lock-keywords-x nil
  "Internal variable used to have extraordinary level font-lock keywords
for the Mail mode.")

(defvar message-font-lock-keywords-x nil
  "Internal variable used to have extraordinary level font-lock keywords
for the Message mode.")

(defvar cmail-font-lock-keywords-x nil
  "Internal variable used to have extraordinary level font-lock keywords
for the Cmail mode.")

(defvar develock-keywords-alist
  '((emacs-lisp-mode lisp-font-lock-keywords-x
		     develock-lisp-font-lock-keywords)
    (lisp-interaction-mode lisp-font-lock-keywords-x
			   develock-lisp-font-lock-keywords)
    (change-log-mode change-log-font-lock-keywords-x
		     develock-change-log-font-lock-keywords)
    (texinfo-mode texinfo-font-lock-keywords-x
		  develock-texinfo-font-lock-keywords)
    (c-mode c-font-lock-keywords-x
	    develock-c-font-lock-keywords)
    (c++-mode c-font-lock-keywords-x
	      develock-c-font-lock-keywords)
    (java-mode java-font-lock-keywords-x
	       develock-java-font-lock-keywords)
    (cperl-mode cperl-font-lock-keywords-x
		develock-cperl-font-lock-keywords)
    (perl-mode perl-font-lock-keywords-x
		develock-perl-font-lock-keywords)
    (mail-mode mail-font-lock-keywords-x
	       develock-mail-font-lock-keywords)
    (message-mode message-font-lock-keywords-x
		  develock-mail-font-lock-keywords)
    (cmail-mail-mode cmail-font-lock-keywords-x
		     develock-mail-font-lock-keywords))
  "*Alist of keyword symbols for major modes.  Each element should be
triple symbols of the following form:

\(major-mode internal-keywords user-defined-keywords)")

(defmacro develock-find-font-lock-defaults (modesym)
  "Get the font-lock defaults based on the major mode."
  (if (featurep 'xemacs)
      `(let ((defaults (or font-lock-defaults
			   (font-lock-find-font-lock-defaults ,modesym))))
	 (if (not (consp (car defaults)))
	     (setcar defaults (list (car defaults))))
	 defaults)
    `(let ((defaults
	     (or font-lock-defaults
		 (cdr (assq ,modesym
			    (symbol-value 'font-lock-defaults-alist))))))
       (cons (if (consp (car defaults))
		 (copy-sequence (car defaults))
	       (list (car defaults)))
	     (copy-sequence (cdr defaults))))))

(defmacro develock-set-defaults (defaults form)
  "Set fontification defaults appropriately for the current `major-mode'."
  (if (featurep 'xemacs)
      `(progn
	 (setq font-lock-defaults-computed nil
	       font-lock-keywords nil)
	 ,form)
    `(progn
       (set (make-local-variable 'font-lock-defaults) ,defaults)
       (setq font-lock-set-defaults nil)
       ,form)))

(defvar develock-inhibit-advice-font-lock-set-defaults nil
  "Internal variable used to inhibit running the advised function
definition of `font-lock-set-defaults' or `font-lock-set-defaults-1'
if it is non-nil.  It will be set to t when the function is called
recursively by itself.  Users should never modify the value.")

(let (current-load-list)
  (eval
   `(defadvice ,(if (featurep 'xemacs)
		    'font-lock-set-defaults-1
		  'font-lock-set-defaults)
      (around energize-font-lock-keywords activate)
      "Advised by Develock.
Energize font-lock keywords for extraordinary level highlighting."
      (if develock-inhibit-advice-font-lock-set-defaults
	  ad-do-it
	(let ((develock-inhibit-advice-font-lock-set-defaults t)
	      (def (cdr (assq major-mode develock-keywords-alist))))
	  (if def
	      (let* ((defaults (develock-find-font-lock-defaults major-mode))
		     (keywords (delq (car def) (car defaults)))
		     (max-decor (nth (1- (length keywords)) keywords))
		     (font-lock-maximum-decoration
		      font-lock-maximum-decoration))
		(setq develock-mode develock-auto-enable)
		(if develock-mode
		    (progn
		      (setq font-lock-maximum-decoration t)
		      (set (car def)
			   (append (if (fboundp max-decor)
				       (funcall max-decor)
				     (symbol-value max-decor))
				   (symbol-value (car (cdr def)))))
		      (setcdr (nthcdr (1- (length keywords)) keywords)
			      (list (car def)))))
		(setcar defaults keywords)
		(develock-set-defaults defaults ad-do-it))
	    ad-do-it))))))

(if (featurep 'xemacs)
    ;; Highlight def* forms even if they aren't started from the
    ;; beginning of the line.
    (let ((def (car (car lisp-font-lock-keywords-1))))
      (if (and (stringp def)
	       (string-match "^\\^\\((\\\\(def\\)" def))
	  (setcar (car lisp-font-lock-keywords-1)
		  (replace-match "^[\t ]*\\1" nil nil def)))))

(defmacro develock-max-column ()
  "Return nil or a number to limit that lines should not be longer than
the value."
  '(let ((value (plist-get develock-max-column-plist major-mode)))
     (cond ((natnump value)
	    value)
	   ((eq 'w value)
	    (1- (window-width)))
	   ((eq t value)
	    (current-fill-column)))))

(eval-and-compile
  (or (fboundp 'kinsoku-bol-p)
      (fboundp 'kinsoku-longer)
      (load "kinsoku" t t)))

(defmacro develock-kinsoku-longer ()
  "Try to leap over the kinsoku characters."
  (cond ((featurep 'xemacs)
	 (if (fboundp 'kinsoku-bol-p)
	     '(while (and (not (eolp))
			  (kinsoku-bol-p))
		(forward-char 1))))
	((fboundp 'kinsoku-longer)
	 '(if enable-kinsoku
	      (while (and (not (eolp))
			  (aref (char-category-set (following-char)) ?>))
		(forward-char 1))))))

(defun develock-find-long-lines (limit)
  "Function used to find a long line between the point and LIMIT.  If a
long line is found, it returns non-nil and sets the `match-data' #0,
#1 and #2 with the beginning and the end positions to be highlighted.
This function is affected by the value of `develock-max-column-plist'."
  (let ((max-column (develock-max-column))
	start end)
    (if (and max-column
	     (progn
	       (while (and (not start)
			   (< (point) limit))
		 (if (or (and (not (eolp))
			      (> (current-column) max-column))
			 (and (or (not (eolp))
				  (zerop (forward-line 1)))
			      (progn
				(while (progn
					 (end-of-line)
					 (and
					  (<= (point) limit)
					  (<= (current-column) max-column)
					  (zerop (forward-line 1)))))
				(> (move-to-column (1+ max-column))
				   max-column))))
		     (progn
		       (if (> (move-to-column max-column) max-column)
			   (backward-char 1))
		       (setq start (min (point) limit))
		       (if (memq major-mode develock-mode-ignore-kinsoku-list)
			   nil
			 (develock-kinsoku-longer)
			 (if (> (point) limit)
			     (goto-char limit))))
		   (forward-line 1)))
	       start)
	     (< start limit))
	(progn
	  (setq start (point)
		end (min (develock-point-at-eol) limit))
	  (or (memq (char-after) '(?\t ?\ ))
	      (skip-chars-backward "0-9A-Za-z"))
	  (if (> (current-column) max-column)
	      (if (> (move-to-column max-column) max-column)
		  (backward-char 1)))
	  (store-match-data (list (point) end
				  (point) start
				  start end))
	  (goto-char end)))))

(eval-when-compile
  (defvar lisp-interaction-mode-hook))

;;;###autoload
(defun develock-mode (arg)
  "Toggle Develock mode.  With arg, turn Develock on if and only if arg
is positive.  In Develock mode, highlighting leading and trailing
whitespaces, long lines and the other garbages."
  (interactive "P")
  (if (assq major-mode develock-keywords-alist)
      (let ((omode develock-mode)
	    already)
	(setq develock-mode (if (numberp arg)
				(> arg 0)
			      (if font-lock-mode
				  (not develock-mode)
				t)))
	(if font-lock-mode
	    (if (eq omode develock-mode)
		(setq already t)
	      (font-lock-mode 0)
	      (let ((develock-auto-enable develock-mode))
		(font-lock-mode 1)))
	  (if develock-mode
	      (let ((develock-auto-enable t))
		(font-lock-mode 1))
	    (setq already t)))
	(message "Develock is %sturned %s"
		 (if already
		     "already "
		   "")
		 (if develock-mode
		     "on"
		   "off")))
    (error "Develock does not support `%s'" major-mode)))

;;;###autoload
(defun turn-on-develock ()
  "Turn on font-lock after splashing the startup screen under XEmacs.
This function is used for `lisp-interaction-mode-hook' exclusively."
  (if (and (boundp 'inhibit-warning-display)
	   (symbol-value 'inhibit-warning-display))
      (add-hook 'pre-command-hook 'turn-on-develock)
    (let ((hooks (cond ((consp lisp-interaction-mode-hook)
			(memq 'turn-on-develock lisp-interaction-mode-hook))
		       ((eq 'turn-on-develock lisp-interaction-mode-hook)
			(setq lisp-interaction-mode-hook
			      '(turn-on-develock))))))
      (if hooks
	  (setcar hooks 'turn-on-font-lock)))
    (if (and (consp pre-command-hook)
	     (memq 'turn-on-develock pre-command-hook))
	(save-excursion
	  (remove-hook 'pre-command-hook 'turn-on-develock)
	  (let ((buffers (buffer-list)))
	    (while buffers
	      (set-buffer (car buffers))
	      (if (eq 'lisp-interaction-mode major-mode)
		  (font-lock-set-defaults))
	      (setq buffers (cdr buffers)))))
      (font-lock-set-defaults))))


;; The following functions are useful to make Lisp codes clean.

(defcustom develock-energize-functions-plist
  '(indent-region t lisp-indent-line t newline-and-indent nil)
  "Plist of function and flag whether to be advised to remove useless
whitespaces."
  :type '(repeat (list :inline t function boolean))
  :set (lambda (symbol value)
	 (set-default symbol value)
	 (let (fn flag)
	   (while value
	     (setq fn (car value)
		   flag (car (cdr value))
		   value (cdr (cdr value)))
	     (condition-case nil
		 (if flag
		     (ad-add-advice fn (plist-get develock-advice-plist fn)
				    'around 'first)
		   (ad-remove-advice fn 'around 'remove-useless-whitespaces))
	       (error)))))
  :group 'develock)

(eval-and-compile
  (or (fboundp 'develock-Orig-lisp-indent-line)
      (defalias 'develock-Orig-lisp-indent-line
	;; The genuine function definition of `lisp-indent-line'.
	(symbol-function (if (fboundp 'ad-Orig-lisp-indent-line)
			     'ad-Orig-lisp-indent-line
			   'lisp-indent-line)))))

(defun develock-lisp-indent-line (&optional whole-exp do-not-move)
  "Internal function used to advise some indent functions."
  (let (pt mod orig)
    (save-excursion
      (end-of-line)
      (setq pt (point))
      (or (zerop (skip-chars-backward "\t "))
	  (delete-region (point) pt))
      (beginning-of-line)
      (or (setq mod (buffer-modified-p))
	  (setq orig (progn
		       (looking-at "[\t ]*")
		       (match-string 0))))
      (insert "\n")
      (backward-char 1)
      (save-restriction
	(widen)
	(develock-Orig-lisp-indent-line whole-exp))
      (if (prog1
	      (bolp)
	    (delete-char 1)
	    (setq pt (point)))
	  (set-buffer-modified-p mod)
	(or (zerop (skip-chars-forward "\t "))
	    (delete-region pt (point)))
	(or mod
	    (progn
	      (beginning-of-line)
	      (if (string-equal orig (progn
				       (looking-at "[\t ]*")
				       (match-string 0)))
		  (set-buffer-modified-p nil))))))
    (or do-not-move
	(goto-char pt))))

(let (current-load-list)
  (defadvice indent-region (around remove-useless-whitespaces
				   (start end column) activate)
    "Advised by Develock.
If Develock is on, remove useless leading and trailing whitespaces in
the Lisp modes.  You can turn off this advice permanently by
customizing the option `develock-energize-functions-plist'."
    (if (and develock-mode font-lock-mode
	     (plist-get develock-energize-functions-plist 'indent-region)
	     (memq major-mode '(emacs-lisp-mode lisp-interaction-mode)))
	(save-excursion
	  ;; Meddle with out of the region.
	  (goto-char end)
	  (while (and (zerop (forward-line 1))
		      (looking-at "[\t ]+$")))
	  (let ((to (point)))
	    (goto-char start)
	    (while (and (zerop (forward-line -1))
			(or (looking-at "[\t ]+$")
			    (progn
			      (forward-line 1)
			      nil))))
	    (save-restriction
	      (if (prog1
		      (zerop (forward-line -1))
		    (narrow-to-region (point) to))
		  (forward-line 1))
	      (while (not (eobp))
		(or (eolp)
		    (progn
		      (develock-lisp-indent-line)
		      (if (and (not (bolp))
			       (eolp))
			  (delete-region (develock-point-at-bol) (point)))))
		(forward-line 1)))))
      ad-do-it))

  (defadvice lisp-indent-line (around remove-useless-whitespaces
				      (&optional whole-exp) activate)
    "Advised by Develock.
If Develock is on, remove useless leading and trailing whitespaces.
You can turn off this advice permanently by customizing the option
`develock-energize-functions-plist'."
    (if (and develock-mode font-lock-mode
	     (plist-get develock-energize-functions-plist 'lisp-indent-line))
	(develock-lisp-indent-line whole-exp)
      ad-do-it))

  (defadvice newline-and-indent (around remove-useless-whitespaces activate)
    "Advised by Develock.
If Develock is on, remove useless leading and trailing whitespaces and
indent appropriately in the Lisp modes.  You can turn off this advice
permanently by customizing the option
`develock-energize-functions-plist'."
    (if (and develock-mode font-lock-mode
	     (plist-get develock-energize-functions-plist 'newline-and-indent)
	     (memq major-mode '(emacs-lisp-mode lisp-interaction-mode)))
	(develock-lisp-indent-line nil t))
    ad-do-it))

(defvar develock-advice-plist nil
  "Internal variable used to keep advice definitions.")

(let ((plist develock-energize-functions-plist)
      fn flag def)
  (while plist
    (setq fn (car plist)
	  flag (car (cdr plist))
	  plist (cdr (cdr plist)))
    (if (setq def (assq 'remove-useless-whitespaces
			(cdr (assq 'around (ad-get-advice-info fn)))))
	(setq develock-advice-plist (plist-put develock-advice-plist fn def)))
    (if (not flag)
	(ad-remove-advice fn 'around 'remove-useless-whitespaces))))

(provide 'develock)

;;; develock.el ends here
